from __future__ import annotations

from functools import lru_cache

import docker
from docker.client import DockerClient

from ._timeouts import DOCKER_REQUEST_TIMEOUT_SECONDS


@lru_cache(maxsize=1)
def get_docker_client() -> DockerClient:
    return docker.from_env(timeout=DOCKER_REQUEST_TIMEOUT_SECONDS)



from __future__ import annotations

import os
import shlex
import subprocess
from pathlib import Path

from docker.errors import ContainerError, DockerException, ImageNotFound

from aicage.docker._client import get_docker_client
from aicage.runtime.env_vars import (
    AICAGE_AGENT_CONFIG_PATH,
    AICAGE_GID,
    AICAGE_UID,
    AICAGE_USER,
    AICAGE_WORKSPACE,
)
from aicage.runtime.run_args import DockerRunArgs


def run_container(args: DockerRunArgs) -> None:
    command = _assemble_docker_run(args)
    subprocess.run(command, check=True)


def print_run_command(args: DockerRunArgs) -> None:
    command = _assemble_docker_run(args)
    print(shlex.join(command))


def run_builder_version_check(image_ref: str, definition_dir: Path) -> subprocess.CompletedProcess[str]:
    command = ["/bin/bash", "/agent/version.sh"]
    volume_src = str(definition_dir.resolve())
    client = get_docker_client()
    try:
        output = client.containers.run(
            image=image_ref,
            command=command,
            volumes={volume_src: {"bind": "/agent", "mode": "ro"}},
            working_dir="/agent",
            remove=True,
            stdout=True,
            stderr=True,
        )
        if isinstance(output, bytes):
            output = output.decode("utf-8", errors="replace")
        return subprocess.CompletedProcess(command, 0, stdout=output, stderr="")
    except ContainerError as exc:
        stdout = _decode_container_output(getattr(exc, "stdout", ""))
        stderr = _decode_container_output(getattr(exc, "stderr", ""))
        return subprocess.CompletedProcess(command, exc.exit_status, stdout=stdout, stderr=stderr)
    except (ImageNotFound, DockerException) as exc:
        return subprocess.CompletedProcess(command, 1, stdout="", stderr=str(exc))


def _decode_container_output(output: object) -> str:
    if isinstance(output, bytes):
        return output.decode("utf-8", errors="replace")
    if isinstance(output, str):
        return output
    return ""


def _resolve_user_ids() -> list[str]:
    env_flags: list[str] = []
    try:
        uid = os.getuid()
        gid = os.getgid()
    except AttributeError:
        uid = gid = None

    user = os.environ.get("USER") or os.environ.get("USERNAME") or "aicage"
    if uid is not None:
        env_flags.extend(["-e", f"{AICAGE_UID}={uid}", "-e", f"{AICAGE_GID}{'='}{gid}"])
    env_flags.extend(["-e", f"{AICAGE_USER}={user}"])
    return env_flags


def _assemble_docker_run(args: DockerRunArgs) -> list[str]:
    cmd: list[str] = ["docker", "run", "--rm", "-it"]
    cmd.extend(_resolve_user_ids())
    cmd.extend(["-e", f"{AICAGE_WORKSPACE}={args.project_path}"])
    if args.agent_path:
        cmd.extend(["-e", f"{AICAGE_AGENT_CONFIG_PATH}={args.agent_path}"])
    for env in args.env:
        cmd.extend(["-e", env])
    cmd.extend(["-v", f"{args.project_path}:/workspace"])
    cmd.extend(["-v", f"{args.project_path}:{args.project_path}"])
    cmd.extend(["-v", f"{args.agent_config_host}:{args.agent_config_mount_container}"])
    for mount in args.mounts:
        suffix = ":ro" if mount.read_only else ""
        cmd.extend(["-v", f"{mount.host_path}:{mount.container_path}{suffix}"])

    if args.merged_docker_args:
        cmd.extend(shlex.split(args.merged_docker_args))

    cmd.append(args.image_ref)
    cmd.extend(args.agent_args)
    return cmd



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config._file_locking import lock_project_config
from aicage.config.config_store import SettingsStore
from aicage.config.context import ConfigContext
from aicage.config.global_config import GlobalConfig
from aicage.config.project_config import AGENT_BASE_KEY, AgentConfig
from aicage.errors import CliError
from aicage.registry.agent_version import AgentVersionChecker
from aicage.registry.image_selection import select_agent_image
from aicage.registry.images_metadata.loader import load_images_metadata
from aicage.registry.images_metadata.models import ImagesMetadata
from aicage.runtime.mounts import resolve_mounts
from aicage.runtime.prompts import prompt_persist_docker_args
from aicage.runtime.run_args import MountSpec


@dataclass(frozen=True)
class RunConfig:
    project_path: Path
    agent: str
    base: str
    image_ref: str
    base_image_ref: str
    extensions: list[str]
    agent_version: str | None
    global_cfg: GlobalConfig
    images_metadata: ImagesMetadata
    project_docker_args: str
    mounts: list[MountSpec]


def load_run_config(agent: str, parsed: ParsedArgs | None = None) -> RunConfig:
    store = SettingsStore()
    project_path = Path.cwd().resolve()
    project_config_path = store.project_config_path(project_path)

    with lock_project_config(project_config_path):
        global_cfg = store.load_global()
        images_metadata = load_images_metadata(global_cfg.local_image_repository)
        project_cfg = store.load_project(project_path)
        context = ConfigContext(
            store=store,
            project_cfg=project_cfg,
            global_cfg=global_cfg,
            images_metadata=images_metadata,
        )
        selection = select_agent_image(agent, context)
        agent_version = _check_agent_version(agent, global_cfg, images_metadata)
        agent_cfg = project_cfg.agents.setdefault(agent, AgentConfig())

        existing_project_docker_args: str = agent_cfg.docker_args

        mounts = resolve_mounts(context, agent, parsed)

        _persist_docker_args(agent_cfg, parsed)
        store.save_project(project_path, project_cfg)

        base = agent_cfg.base or global_cfg.agents.get(agent, {}).get(AGENT_BASE_KEY)
        if base is None:
            raise CliError(f"Base selection is missing for agent '{agent}'.")

        return RunConfig(
            project_path=project_path,
            agent=agent,
            base=base,
            image_ref=selection.image_ref,
            base_image_ref=selection.base_image_ref,
            extensions=list(selection.extensions),
            agent_version=agent_version,
            global_cfg=global_cfg,
            images_metadata=images_metadata,
            project_docker_args=existing_project_docker_args,
            mounts=mounts,
        )


def _persist_docker_args(agent_cfg: AgentConfig, parsed: ParsedArgs | None) -> None:
    if parsed is None or not parsed.docker_args:
        return
    existing = agent_cfg.docker_args
    if existing == parsed.docker_args:
        return

    if prompt_persist_docker_args(parsed.docker_args, existing):
        agent_cfg.docker_args = parsed.docker_args


def _check_agent_version(
    agent: str,
    global_cfg: GlobalConfig,
    images_metadata: ImagesMetadata,
) -> str | None:
    agent_metadata = images_metadata.agents[agent]
    definition_dir = agent_metadata.local_definition_dir
    if definition_dir is None:
        return None
    checker = AgentVersionChecker(global_cfg)
    return checker.get_version(agent, agent_metadata, definition_dir)



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import portalocker

from aicage.config.errors import ConfigError

_LOCK_TIMEOUT_SECONDS = 30


@contextmanager
def lock_project_config(project_config_path: Path) -> Iterator[None]:
    try:
        with _lock_file(project_config_path):
            yield
    except portalocker.exceptions.LockException as exc:  # pragma: no cover - rare file lock failure
        raise ConfigError(f"Failed to lock project configuration file: {exc}") from exc


def _lock_file(path: Path) -> portalocker.Lock:
    path.parent.mkdir(parents=True, exist_ok=True)
    return portalocker.Lock(str(path), timeout=_LOCK_TIMEOUT_SECONDS, mode="a+")



from pathlib import Path

from aicage.config.project_config import AgentConfig
from aicage.runtime.prompts import prompt_mount_git_config
from aicage.runtime.run_args import MountSpec

from ._exec import capture_stdout

_GITCONFIG_MOUNT = Path("/aicage/host/gitconfig")


def _resolve_git_config_path() -> Path | None:
    stdout = capture_stdout(["git", "config", "--global", "--show-origin", "--list"])
    if not stdout:
        return None
    for line in stdout.splitlines():
        if not line.startswith("file:"):
            continue
        parts = line[5:].split()
        if not parts:
            continue
        return Path(parts[0]).expanduser()
    return None


def resolve_git_config_mount(agent_cfg: AgentConfig) -> list[MountSpec]:
    git_config = _resolve_git_config_path()
    if not git_config or not git_config.exists():
        return []

    mounts_cfg = agent_cfg.mounts
    pref = mounts_cfg.gitconfig
    if pref is None:
        pref = prompt_mount_git_config(git_config)
        mounts_cfg.gitconfig = pref

    if pref:
        return [MountSpec(host_path=git_config, container_path=_GITCONFIG_MOUNT)]
    return []



from pathlib import Path

from aicage._logging import get_logger

from ._tty import ensure_tty_for_prompt


def prompt_yes_no(question: str, default: bool = False) -> bool:
    ensure_tty_for_prompt()
    suffix = "[Y/n]" if default else "[y/N]"
    response = input(f"{question} {suffix} ").strip().lower()
    if not response:
        choice = default
    else:
        choice = response in {"y", "yes"}
    get_logger().info("Prompt yes/no '%s' -> %s", question, choice)
    return choice


def prompt_persist_entrypoint(entrypoint_path: Path) -> bool:
    return prompt_yes_no(f"Persist entrypoint '{entrypoint_path}' for this project?", default=True)


def prompt_persist_docker_socket() -> bool:
    return prompt_yes_no("Persist mounting the Docker socket for this project?", default=True)


def prompt_mount_git_config(git_config: Path) -> bool:
    question = f"Mount Git config from '{git_config}' so Git uses your usual name/email?"
    return prompt_yes_no(question, default=True)


def prompt_mount_gpg_keys(gpg_home: Path) -> bool:
    question = f"Mount GnuPG keys from '{gpg_home}' so Git signing works like on your host?"
    return prompt_yes_no(question, default=True)


def prompt_mount_ssh_keys(ssh_dir: Path) -> bool:
    question = f"Mount SSH keys from '{ssh_dir}' so Git signing works like on your host?"
    return prompt_yes_no(question, default=True)


def prompt_persist_docker_args(new_args: str, existing_args: str | None) -> bool:
    if existing_args:
        question = f"Persist docker run args '{new_args}' for this project (replacing '{existing_args}')?"
    else:
        question = f"Persist docker run args '{new_args}' for this project?"
    return prompt_yes_no(question, default=True)



from __future__ import annotations


def local_image_ref(local_image_repository: str, agent: str, base: str) -> str:
    tag = f"{agent}-{base}".lower().replace("/", "-")
    return f"{local_image_repository}:{tag}"



from __future__ import annotations

from docker.errors import DockerException, ImageNotFound

from ._client import get_docker_client
from .types import ImageRefRepository


def local_image_exists(image_ref: str) -> bool:
    client = get_docker_client()
    try:
        client.images.get(image_ref)
    except ImageNotFound:
        return False
    return True


def get_local_repo_digest(image: ImageRefRepository) -> str | None:
    return get_local_repo_digest_for_repo(image.image_ref, image.repository)


def get_local_repo_digest_for_repo(image_ref: str, repository: str) -> str | None:
    try:
        client = get_docker_client()
        image = client.images.get(image_ref)
    except (ImageNotFound, DockerException):
        return None

    repo_digests = image.attrs.get("RepoDigests")
    if not isinstance(repo_digests, list):
        return None

    for entry in repo_digests:
        if not isinstance(entry, str):
            continue
        repo, sep, digest = entry.partition("@")
        if sep and repo == repository and digest:
            return digest

    return None


def get_local_rootfs_layers(image_ref: str) -> list[str] | None:
    try:
        client = get_docker_client()
        image = client.images.get(image_ref)
    except (ImageNotFound, DockerException):
        return None

    rootfs = image.attrs.get("RootFS")
    if not isinstance(rootfs, dict):
        return None
    layers = rootfs.get("Layers")
    if not isinstance(layers, list):
        return None
    filtered = [layer for layer in layers if isinstance(layer, str)]
    if not filtered:
        return None
    return filtered



from dataclasses import dataclass, field
from pathlib import Path


@dataclass
class MountSpec:
    host_path: Path
    container_path: Path
    read_only: bool = False


@dataclass
class DockerRunArgs:
    image_ref: str
    project_path: Path
    agent_config_host: Path
    agent_config_mount_container: Path
    merged_docker_args: str
    agent_args: list[str]
    agent_path: str | None = None
    env: list[str] = field(default_factory=list)
    mounts: list[MountSpec] = field(default_factory=list)


def merge_docker_args(*args: str) -> str:
    return " ".join(part for part in args if part).strip()



from __future__ import annotations

from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.runtime.run_args import MountSpec

from ._docker_socket import resolve_docker_socket_mount
from ._entrypoint import resolve_entrypoint_mount
from ._git_config import resolve_git_config_mount
from ._gpg import resolve_gpg_mount
from ._ssh_keys import resolve_ssh_mount


def resolve_mounts(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> list[MountSpec]:
    agent_cfg = context.project_cfg.agents.setdefault(agent, AgentConfig())

    git_mounts = resolve_git_config_mount(agent_cfg)
    project_path = Path(context.project_cfg.path)
    ssh_mounts = resolve_ssh_mount(project_path, agent_cfg)
    gpg_mounts = resolve_gpg_mount(project_path, agent_cfg)
    entrypoint_mounts = resolve_entrypoint_mount(
        agent_cfg,
        parsed.entrypoint if parsed else None,
    )
    docker_mounts = resolve_docker_socket_mount(
        agent_cfg,
        parsed.docker_socket if parsed else False,
    )

    mounts: list[MountSpec] = []
    mounts.extend(git_mounts)
    mounts.extend(ssh_mounts)
    mounts.extend(gpg_mounts)
    mounts.extend(entrypoint_mounts)
    mounts.extend(docker_mounts)
    return mounts



from dataclasses import dataclass, field
from typing import Any

from .errors import ConfigError

_IMAGE_REGISTRY_KEY: str = "image_registry"
_IMAGE_REGISTRY_API_URL_KEY: str = "image_registry_api_url"
_IMAGE_REGISTRY_API_TOKEN_URL_KEY: str = "image_registry_api_token_url"
_IMAGE_REPOSITORY_KEY: str = "image_repository"
_IMAGE_BASE_REPOSITORY_KEY: str = "image_base_repository"
_DEFAULT_IMAGE_BASE_KEY: str = "default_image_base"
_VERSION_CHECK_IMAGE_KEY: str = "version_check_image"
_LOCAL_IMAGE_REPOSITORY_KEY: str = "local_image_repository"
_GLOBAL_AGENTS_KEY: str = "agents"


@dataclass
class GlobalConfig:
    image_registry: str
    image_registry_api_url: str
    image_registry_api_token_url: str
    image_repository: str
    image_base_repository: str
    default_image_base: str
    version_check_image: str
    local_image_repository: str
    agents: dict[str, dict[str, Any]] = field(default_factory=dict)

    @classmethod
    def from_mapping(cls, data: dict[str, Any]) -> "GlobalConfig":
        required = (
            _IMAGE_REGISTRY_KEY,
            _IMAGE_REGISTRY_API_URL_KEY,
            _IMAGE_REGISTRY_API_TOKEN_URL_KEY,
            _IMAGE_REPOSITORY_KEY,
            _IMAGE_BASE_REPOSITORY_KEY,
            _DEFAULT_IMAGE_BASE_KEY,
            _VERSION_CHECK_IMAGE_KEY,
            _LOCAL_IMAGE_REPOSITORY_KEY,
        )
        missing = [key for key in required if key not in data]
        if missing:
            raise ConfigError(f"Missing required config values: {', '.join(missing)}.")
        return cls(
            image_registry=data[_IMAGE_REGISTRY_KEY],
            image_registry_api_url=data[_IMAGE_REGISTRY_API_URL_KEY],
            image_registry_api_token_url=data[_IMAGE_REGISTRY_API_TOKEN_URL_KEY],
            image_repository=data[_IMAGE_REPOSITORY_KEY],
            image_base_repository=data[_IMAGE_BASE_REPOSITORY_KEY],
            default_image_base=data[_DEFAULT_IMAGE_BASE_KEY],
            version_check_image=data[_VERSION_CHECK_IMAGE_KEY],
            local_image_repository=data[_LOCAL_IMAGE_REPOSITORY_KEY],
            agents=data.get(_GLOBAL_AGENTS_KEY, {}) or {},
        )

    def to_mapping(self) -> dict[str, Any]:
        return {
            _IMAGE_REGISTRY_KEY: self.image_registry,
            _IMAGE_REGISTRY_API_URL_KEY: self.image_registry_api_url,
            _IMAGE_REGISTRY_API_TOKEN_URL_KEY: self.image_registry_api_token_url,
            _IMAGE_REPOSITORY_KEY: self.image_repository,
            _IMAGE_BASE_REPOSITORY_KEY: self.image_base_repository,
            _DEFAULT_IMAGE_BASE_KEY: self.default_image_base,
            _VERSION_CHECK_IMAGE_KEY: self.version_check_image,
            _LOCAL_IMAGE_REPOSITORY_KEY: self.local_image_repository,
            _GLOBAL_AGENTS_KEY: self.agents,
        }



from __future__ import annotations

from aicage._logging import get_logger
from aicage.config.runtime_config import RunConfig
from aicage.docker.pull import run_pull
from aicage.registry._logs import pull_log_path
from aicage.registry._pull_decision import decide_pull


def pull_image(run_config: RunConfig) -> None:
    logger = get_logger()
    decision = decide_pull(run_config)
    if not decision.should_pull:
        logger.info("Image pull not required for %s", run_config.image_ref)
        return

    log_path = pull_log_path(run_config.image_ref)
    run_pull(run_config.image_ref, log_path)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml

from aicage.errors import CliError
from aicage.paths import CUSTOM_EXTENSION_DEFINITION_FILES, DEFAULT_CUSTOM_EXTENSIONS_DIR
from aicage.registry._hashing import HashWriter, new_hasher

_EXTENSION_NAME_KEY: str = "name"
_EXTENSION_DESCRIPTION_KEY: str = "description"
_SCRIPTS_DIRNAME: str = "scripts"
_DOCKERFILE_NAME: str = "Dockerfile"


@dataclass(frozen=True)
class ExtensionMetadata:
    extension_id: str
    name: str
    description: str
    directory: Path
    scripts_dir: Path
    dockerfile_path: Path | None


def load_extensions() -> dict[str, ExtensionMetadata]:
    extensions_dir = DEFAULT_CUSTOM_EXTENSIONS_DIR.expanduser()
    if not extensions_dir.is_dir():
        return {}
    extensions: dict[str, ExtensionMetadata] = {}
    for entry in sorted(extensions_dir.iterdir()):
        if not entry.is_dir():
            continue
        extension_id = entry.name
        definition_path = _find_extension_definition(entry)
        mapping = _load_yaml(definition_path)
        _expect_keys(
            mapping,
            required={_EXTENSION_NAME_KEY, _EXTENSION_DESCRIPTION_KEY},
            optional=set(),
            context=f"extension metadata at {definition_path}",
        )
        scripts_dir = entry / _SCRIPTS_DIRNAME
        if not scripts_dir.is_dir():
            raise CliError(f"Extension '{extension_id}' is missing scripts/ directory.")
        dockerfile_path = entry / _DOCKERFILE_NAME
        extensions[extension_id] = ExtensionMetadata(
            extension_id=extension_id,
            name=_expect_string(mapping.get(_EXTENSION_NAME_KEY), _EXTENSION_NAME_KEY),
            description=_expect_string(mapping.get(_EXTENSION_DESCRIPTION_KEY), _EXTENSION_DESCRIPTION_KEY),
            directory=entry,
            scripts_dir=scripts_dir,
            dockerfile_path=dockerfile_path if dockerfile_path.is_file() else None,
        )
    return extensions


def extension_hash(extension: ExtensionMetadata) -> str:
    digest = new_hasher()
    definition_path = _find_extension_definition(extension.directory)
    _update_hash(digest, definition_path)
    if extension.dockerfile_path is not None:
        _update_hash(digest, extension.dockerfile_path)
    for script in sorted(extension.scripts_dir.glob("*.sh")):
        if script.is_file():
            _update_hash(digest, script)
    return digest.hexdigest()


def _find_extension_definition(extension_dir: Path) -> Path:
    for filename in CUSTOM_EXTENSION_DEFINITION_FILES:
        candidate = extension_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(CUSTOM_EXTENSION_DEFINITION_FILES)
    raise CliError(f"Extension '{extension_dir.name}' is missing {expected}.")


def _update_hash(digest: HashWriter, path: Path) -> None:
    digest.update(path.name.encode("utf-8"))
    digest.update(path.read_bytes())


def _load_yaml(path: Path) -> dict[str, Any]:
    try:
        payload = path.read_text(encoding="utf-8")
        data = yaml.safe_load(payload) or {}
    except (OSError, yaml.YAMLError) as exc:
        raise CliError(f"Failed to read extension metadata from {path}: {exc}") from exc
    if not isinstance(data, dict):
        raise CliError(f"Extension metadata at {path} must be a mapping.")
    return data


def _expect_string(value: Any, context: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise CliError(f"{context} must be a non-empty string.")
    return value


def _expect_keys(
    mapping: dict[str, Any],
    required: set[str],
    optional: set[str],
    context: str,
) -> None:
    missing = sorted(required - set(mapping))
    if missing:
        raise CliError(f"{context} missing required keys: {', '.join(missing)}.")
    unknown = sorted(set(mapping) - required - optional)
    if unknown:
        raise CliError(f"{context} contains unsupported keys: {', '.join(unknown)}.")



from __future__ import annotations

import logging
import os
from pathlib import Path

_LOGGER_NAME = "aicage"
_LOG_LEVEL_ENV = "AICAGE_LOG_LEVEL"
_DEFAULT_LEVEL = "INFO"
_LOG_PATH = "~/.aicage/logs/aicage.log"


def get_logger() -> logging.Logger:
    logger = logging.getLogger(_LOGGER_NAME)
    if logger.handlers:
        return logger

    log_path = Path(os.path.expanduser(_LOG_PATH))
    log_path.parent.mkdir(parents=True, exist_ok=True)

    level = _resolve_level(os.getenv(_LOG_LEVEL_ENV, _DEFAULT_LEVEL))
    logger.setLevel(level)
    logger.propagate = False

    handler = logging.FileHandler(log_path, encoding="utf-8")
    handler.setLevel(level)
    handler.setFormatter(
        logging.Formatter("%(asctime)s %(levelname)s %(name)s %(message)s")
    )
    logger.addHandler(handler)
    return logger


def _resolve_level(raw_level: str) -> int:
    normalized = raw_level.strip().upper()
    level = logging.getLevelName(normalized)
    if isinstance(level, int):
        return level
    return logging.INFO



from aicage.config.context import ConfigContext
from aicage.registry._image_refs import local_image_ref
from aicage.registry.images_metadata.models import AgentMetadata


def base_image_ref(
    agent_metadata: AgentMetadata,
    agent: str,
    base: str,
    context: ConfigContext,
) -> str:
    if agent_metadata.local_definition_dir is not None:
        return local_image_ref(context.global_cfg.local_image_repository, agent, base)
    return agent_metadata.valid_bases[base]



from pathlib import Path

from aicage.config.project_config import AgentConfig
from aicage.runtime.prompts import prompt_mount_ssh_keys
from aicage.runtime.run_args import MountSpec

from ._signing import is_commit_signing_enabled, resolve_signing_format

_SSH_MOUNT = Path("/aicage/host/ssh")


def _default_ssh_dir() -> Path:
    return Path.home() / ".ssh"


def resolve_ssh_mount(project_path: Path, agent_cfg: AgentConfig) -> list[MountSpec]:
    if not is_commit_signing_enabled(project_path):
        return []
    if resolve_signing_format(project_path) != "ssh":
        return []

    ssh_dir = _default_ssh_dir()
    if not ssh_dir.exists():
        return []

    mounts_cfg = agent_cfg.mounts
    pref = mounts_cfg.ssh
    if pref is None:
        pref = prompt_mount_ssh_keys(ssh_dir)
        mounts_cfg.ssh = pref

    if pref:
        return [MountSpec(host_path=ssh_dir, container_path=_SSH_MOUNT)]
    return []



from __future__ import annotations

from aicage.config.runtime_config import RunConfig
from aicage.docker.build import run_extended_build
from aicage.errors import CliError
from aicage.registry._hashing import new_hasher
from aicage.registry.extensions import ExtensionMetadata, extension_hash, load_extensions
from aicage.registry.image_pull import pull_image
from aicage.registry.local_build.ensure_local_image import ensure_local_image

from ._extended_plan import should_build_extended
from ._extended_store import ExtendedBuildRecord, ExtendedBuildStore
from ._logs import build_log_path_for_image
from ._plan import now_iso


def ensure_extended_image(run_config: RunConfig) -> None:
    if not run_config.extensions:
        raise CliError("No extensions selected for extended image build.")

    extensions = load_extensions()
    resolved = _resolve_extensions(run_config.extensions, extensions)
    base_run_config = _base_run_config(run_config)

    agent_metadata = run_config.images_metadata.agents[run_config.agent]
    if agent_metadata.local_definition_dir is None:
        pull_image(base_run_config)
    else:
        ensure_local_image(base_run_config)

    combined_hash = _combined_extension_hash(resolved)
    store = ExtendedBuildStore()
    record = store.load(run_config.image_ref)
    needs_build = should_build_extended(
        run_config=run_config,
        record=record,
        base_image_ref=run_config.base_image_ref,
        extension_hash=combined_hash,
    )
    if not needs_build:
        return

    log_path = build_log_path_for_image(run_config.image_ref)
    run_extended_build(
        run_config=run_config,
        base_image_ref=run_config.base_image_ref,
        extensions=resolved,
        log_path=log_path,
    )
    store.save(
        ExtendedBuildRecord(
            agent=run_config.agent,
            base=run_config.base,
            image_ref=run_config.image_ref,
            extensions=list(run_config.extensions),
            extension_hash=combined_hash,
            base_image=run_config.base_image_ref,
            built_at=now_iso(),
        )
    )


def _resolve_extensions(
    extension_ids: list[str],
    extensions: dict[str, ExtensionMetadata],
) -> list[ExtensionMetadata]:
    missing = [ext for ext in extension_ids if ext not in extensions]
    if missing:
        raise CliError(f"Missing extensions: {', '.join(sorted(missing))}.")
    return [extensions[ext] for ext in extension_ids]


def _combined_extension_hash(extensions: list[ExtensionMetadata]) -> str:
    digest = new_hasher()
    for extension in extensions:
        digest.update(extension.extension_id.encode("utf-8"))
        digest.update(extension_hash(extension).encode("utf-8"))
    return digest.hexdigest()


def _base_run_config(run_config: RunConfig) -> RunConfig:
    return RunConfig(
        project_path=run_config.project_path,
        agent=run_config.agent,
        base=run_config.base,
        image_ref=run_config.base_image_ref,
        base_image_ref=run_config.base_image_ref,
        extensions=[],
        agent_version=run_config.agent_version,
        global_cfg=run_config.global_cfg,
        images_metadata=run_config.images_metadata,
        project_docker_args=run_config.project_docker_args,
        mounts=run_config.mounts,
    )



from __future__ import annotations

import subprocess
from logging import Logger
from pathlib import Path

from aicage._logging import get_logger
from aicage.config.resources import find_packaged_path
from aicage.config.runtime_config import RunConfig
from aicage.errors import CliError
from aicage.registry.extensions import ExtensionMetadata


def run_build(
    run_config: RunConfig,
    base_image_ref: str,
    log_path: Path,
) -> None:
    logger = get_logger()
    log_path.parent.mkdir(parents=True, exist_ok=True)
    print(f"[aicage] Building local image {run_config.image_ref} (logs: {log_path})...")
    logger.info("Building local image %s (logs: %s)", run_config.image_ref, log_path)

    dockerfile_path = find_packaged_path("agent-build/Dockerfile")
    build_root = _build_context_dir(run_config, dockerfile_path)
    command = [
        "docker",
        "build",
        "--no-cache",
        "--file",
        str(dockerfile_path),
        "--build-arg",
        f"BASE_IMAGE={base_image_ref}",
        "--build-arg",
        f"AGENT={run_config.agent}",
        "--tag",
        run_config.image_ref,
        str(build_root),
    ]
    with log_path.open("w", encoding="utf-8") as log_handle:
        result = subprocess.run(command, check=False, stdout=log_handle, stderr=subprocess.STDOUT)
    if result.returncode != 0:
        logger.error("Local image build failed for %s (logs: %s)", run_config.image_ref, log_path)
        raise CliError(
            f"Local image build failed for {run_config.image_ref}. See log at {log_path}."
        )

    logger.info("Local image build succeeded for %s", run_config.image_ref)


def run_extended_build(
    run_config: RunConfig,
    base_image_ref: str,
    extensions: list[ExtensionMetadata],
    log_path: Path,
) -> None:
    logger = get_logger()
    log_path.parent.mkdir(parents=True, exist_ok=True)
    print(f"[aicage] Building extended image {run_config.image_ref} (logs: {log_path})...")
    logger.info("Building extended image %s (logs: %s)", run_config.image_ref, log_path)

    dockerfile_builtin = find_packaged_path("extension-build/Dockerfile")
    current_image_ref = base_image_ref
    intermediate_refs: list[str] = []
    with log_path.open("w", encoding="utf-8") as log_handle:
        for idx, extension in enumerate(extensions):
            target_ref = (
                run_config.image_ref
                if idx == len(extensions) - 1
                else _intermediate_image_ref(run_config, extension, idx)
            )
            if target_ref != run_config.image_ref:
                intermediate_refs.append(target_ref)
            dockerfile_path = extension.dockerfile_path or dockerfile_builtin
            command = [
                "docker",
                "build",
                "--no-cache",
                "--file",
                str(dockerfile_path),
                "--build-arg",
                f"BASE_IMAGE={current_image_ref}",
                "--build-arg",
                f"EXTENSION={extension.extension_id}",
                "--tag",
                target_ref,
                str(extension.directory),
            ]
            result = subprocess.run(command, check=False, stdout=log_handle, stderr=subprocess.STDOUT)
            if result.returncode != 0:
                logger.error(
                    "Extended image build failed for %s (logs: %s)",
                    run_config.image_ref,
                    log_path,
                )
                raise CliError(
                    f"Extended image build failed for {run_config.image_ref}. See log at {log_path}."
                )
            current_image_ref = target_ref
    _cleanup_intermediate_images(intermediate_refs, logger)
    logger.info("Extended image build succeeded for %s", run_config.image_ref)


def _build_context_dir(run_config: RunConfig, dockerfile_path: Path) -> Path:
    agent_metadata = run_config.images_metadata.agents[run_config.agent]
    local_definition_dir = agent_metadata.local_definition_dir
    if local_definition_dir is None:
        return dockerfile_path.parent
    if local_definition_dir.is_relative_to(dockerfile_path.parent):
        return dockerfile_path.parent
    return local_definition_dir.parent.parent


def _intermediate_image_ref(run_config: RunConfig, extension: ExtensionMetadata, idx: int) -> str:
    repository, _ = _parse_image_ref(run_config.image_ref)
    tag = f"tmp-{run_config.agent}-{run_config.base}-{idx + 1}-{extension.extension_id}"
    tag = tag.lower().replace("/", "-")
    return f"{repository}:{tag}"


def _parse_image_ref(image_ref: str) -> tuple[str, str]:
    repository, sep, tag = image_ref.rpartition(":")
    if not sep:
        raise CliError(f"Image ref '{image_ref}' is missing a tag.")
    return repository, tag


def _cleanup_intermediate_images(intermediate_refs: list[str], logger: Logger) -> None:
    for image_ref in intermediate_refs:
        result = subprocess.run(
            ["docker", "image", "rm", "-f", image_ref],
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if result.returncode != 0:
            logger.info("Failed to remove intermediate image %s", image_ref)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml

from aicage.config.resources import find_packaged_path
from aicage.errors import CliError

_AICAGE_IMAGE_KEY: str = "aicage-image"
_AICAGE_IMAGE_BASE_KEY: str = "aicage-image-base"
_BASES_KEY: str = "bases"
_AGENT_KEY: str = "agent"

_VERSION_KEY: str = "version"

_ROOT_IMAGE_KEY: str = "root_image"
_BASE_IMAGE_DISTRO_KEY: str = "base_image_distro"
_BASE_IMAGE_DESCRIPTION_KEY: str = "base_image_description"
_OS_INSTALLER_KEY: str = "os_installer"
_TEST_SUITE_KEY: str = "test_suite"

AGENT_PATH_KEY: str = "agent_path"
AGENT_FULL_NAME_KEY: str = "agent_full_name"
AGENT_HOMEPAGE_KEY: str = "agent_homepage"
BUILD_LOCAL_KEY: str = "build_local"
_VALID_BASES_KEY: str = "valid_bases"
BASE_EXCLUDE_KEY: str = "base_exclude"
BASE_DISTRO_EXCLUDE_KEY: str = "base_distro_exclude"


@dataclass(frozen=True)
class _ImageReleaseInfo:
    version: str


@dataclass(frozen=True)
class _BaseMetadata:
    root_image: str
    base_image_distro: str
    base_image_description: str
    os_installer: str
    test_suite: str


@dataclass(frozen=True)
class AgentMetadata:
    agent_path: str
    agent_full_name: str
    agent_homepage: str
    valid_bases: dict[str, str]
    base_exclude: list[str] | None = None
    base_distro_exclude: list[str] | None = None
    local_definition_dir: Path | None = None


@dataclass(frozen=True)
class ImagesMetadata:
    aicage_image: _ImageReleaseInfo
    aicage_image_base: _ImageReleaseInfo
    bases: dict[str, _BaseMetadata]
    agents: dict[str, AgentMetadata]

    @classmethod
    def from_yaml(cls, payload: str) -> ImagesMetadata:
        try:
            data = yaml.safe_load(payload) or {}
        except yaml.YAMLError as exc:
            raise CliError(f"Invalid images metadata YAML: {exc}") from exc
        if not isinstance(data, dict):
            raise CliError("Images metadata YAML must be a mapping at the top level.")
        return cls.from_mapping(data)

    @classmethod
    def from_mapping(cls, data: dict[str, Any]) -> ImagesMetadata:
        _expect_keys(
            data,
            required={_AICAGE_IMAGE_KEY, _AICAGE_IMAGE_BASE_KEY, _BASES_KEY, _AGENT_KEY},
            optional=set(),
            context="images metadata",
        )
        aicage_image = _parse_release_info(data[_AICAGE_IMAGE_KEY], _AICAGE_IMAGE_KEY)
        aicage_image_base = _parse_release_info(data[_AICAGE_IMAGE_BASE_KEY], _AICAGE_IMAGE_BASE_KEY)
        bases = _parse_bases(data[_BASES_KEY])
        agents = _parse_agents(data[_AGENT_KEY])
        return cls(
            aicage_image=aicage_image,
            aicage_image_base=aicage_image_base,
            bases=bases,
            agents=agents,
        )


def _parse_release_info(value: Any, context: str) -> _ImageReleaseInfo:
    mapping = _expect_mapping(value, context)
    _expect_keys(mapping, required={_VERSION_KEY}, optional=set(), context=context)
    return _ImageReleaseInfo(version=_expect_string(mapping.get(_VERSION_KEY), f"{context}.{_VERSION_KEY}"))


def _parse_bases(value: Any) -> dict[str, _BaseMetadata]:
    mapping = _expect_mapping(value, _BASES_KEY)
    bases: dict[str, _BaseMetadata] = {}
    for name, base_value in mapping.items():
        if not isinstance(name, str):
            raise CliError("Images metadata base keys must be strings.")
        base_mapping = _expect_mapping(base_value, f"{_BASES_KEY}.{name}")
        _expect_keys(
            base_mapping,
            required={
                _ROOT_IMAGE_KEY,
                _BASE_IMAGE_DISTRO_KEY,
                _BASE_IMAGE_DESCRIPTION_KEY,
                _OS_INSTALLER_KEY,
                _TEST_SUITE_KEY,
            },
            optional=set(),
            context=f"{_BASES_KEY}.{name}",
        )
        bases[name] = _BaseMetadata(
            root_image=_expect_string(
                base_mapping.get(_ROOT_IMAGE_KEY),
                f"{_BASES_KEY}.{name}.{_ROOT_IMAGE_KEY}",
            ),
            base_image_distro=_expect_string(
                base_mapping.get(_BASE_IMAGE_DISTRO_KEY),
                f"{_BASES_KEY}.{name}.{_BASE_IMAGE_DISTRO_KEY}",
            ),
            base_image_description=_expect_string(
                base_mapping.get(_BASE_IMAGE_DESCRIPTION_KEY),
                f"{_BASES_KEY}.{name}.{_BASE_IMAGE_DESCRIPTION_KEY}",
            ),
            os_installer=_expect_string(
                base_mapping.get(_OS_INSTALLER_KEY),
                f"{_BASES_KEY}.{name}.{_OS_INSTALLER_KEY}",
            ),
            test_suite=_expect_string(
                base_mapping.get(_TEST_SUITE_KEY),
                f"{_BASES_KEY}.{name}.{_TEST_SUITE_KEY}",
            ),
        )
    return bases


def _parse_agents(value: Any) -> dict[str, AgentMetadata]:
    mapping = _expect_mapping(value, _AGENT_KEY)
    agents: dict[str, AgentMetadata] = {}
    for name, agent_value in mapping.items():
        if not isinstance(name, str):
            raise CliError("Images metadata agent keys must be strings.")
        agent_mapping = _expect_mapping(agent_value, f"{_AGENT_KEY}.{name}")
        _expect_keys(
            agent_mapping,
            required={
                AGENT_PATH_KEY,
                AGENT_FULL_NAME_KEY,
                AGENT_HOMEPAGE_KEY,
                BUILD_LOCAL_KEY,
                _VALID_BASES_KEY,
            },
            optional={BASE_EXCLUDE_KEY, BASE_DISTRO_EXCLUDE_KEY},
            context=f"{_AGENT_KEY}.{name}",
        )
        agents[name] = AgentMetadata(
            agent_path=_expect_string(
                agent_mapping.get(AGENT_PATH_KEY), f"{_AGENT_KEY}.{name}.{AGENT_PATH_KEY}"
            ),
            agent_full_name=_expect_string(
                agent_mapping.get(AGENT_FULL_NAME_KEY),
                f"{_AGENT_KEY}.{name}.{AGENT_FULL_NAME_KEY}",
            ),
            agent_homepage=_expect_string(
                agent_mapping.get(AGENT_HOMEPAGE_KEY),
                f"{_AGENT_KEY}.{name}.{AGENT_HOMEPAGE_KEY}",
            ),
            local_definition_dir=_local_definition_dir(
                name,
                _expect_bool(
                    agent_mapping.get(BUILD_LOCAL_KEY), f"{_AGENT_KEY}.{name}.{BUILD_LOCAL_KEY}"
                ),
            ),
            valid_bases=_expect_str_mapping(
                agent_mapping.get(_VALID_BASES_KEY), f"{_AGENT_KEY}.{name}.{_VALID_BASES_KEY}"
            ),
            base_exclude=_maybe_str_list(
                agent_mapping.get(BASE_EXCLUDE_KEY), f"{_AGENT_KEY}.{name}.{BASE_EXCLUDE_KEY}"
            ),
            base_distro_exclude=_maybe_str_list(
                agent_mapping.get(BASE_DISTRO_EXCLUDE_KEY),
                f"{_AGENT_KEY}.{name}.{BASE_DISTRO_EXCLUDE_KEY}",
            ),
        )
    return agents


def _local_definition_dir(agent_name: str, build_local: bool) -> Path | None:
    if not build_local:
        return None
    dockerfile = find_packaged_path("agent-build/Dockerfile")
    return dockerfile.parent / "agents" / agent_name


def _expect_mapping(value: Any, context: str) -> dict[str, Any]:
    if not isinstance(value, dict):
        raise CliError(f"{context} must be a mapping.")
    return value


def _expect_string(value: Any, context: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise CliError(f"{context} must be a non-empty string.")
    return value


def _expect_bool(value: Any, context: str) -> bool:
    if not isinstance(value, bool):
        raise CliError(f"{context} must be a boolean.")
    return value


def _expect_str_list(value: Any, context: str) -> list[str]:
    if not isinstance(value, list):
        raise CliError(f"{context} must be a list.")
    items: list[str] = []
    for item in value:
        if not isinstance(item, str) or not item.strip():
            raise CliError(f"{context} must contain non-empty strings.")
        items.append(item)
    return items


def _expect_str_mapping(value: Any, context: str) -> dict[str, str]:
    mapping = _expect_mapping(value, context)
    items: dict[str, str] = {}
    for key, item in mapping.items():
        if not isinstance(key, str) or not key.strip():
            raise CliError(f"{context} must contain non-empty string keys.")
        if not isinstance(item, str) or not item.strip():
            raise CliError(f"{context} must contain non-empty string values.")
        items[key] = item
    return items


def _maybe_str_list(value: Any, context: str) -> list[str] | None:
    if value is None:
        return None
    return _expect_str_list(value, context)


def _expect_keys(
    mapping: dict[str, Any],
    required: set[str],
    optional: set[str],
    context: str,
) -> None:
    missing = sorted(required - set(mapping))
    if missing:
        raise CliError(f"{context} missing required keys: {', '.join(missing)}.")
    unknown = sorted(set(mapping) - required - optional)
    if unknown:
        raise CliError(f"{context} contains unsupported keys: {', '.join(unknown)}.")



from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path

import yaml

from aicage.paths import DEFAULT_VERSION_CHECK_STATE_DIR

_AGENT_KEY: str = "agent"
_VERSION_KEY: str = "version"
_CHECKED_AT_KEY: str = "checked_at"


class VersionCheckStore:
    def __init__(self, base_dir: Path | None = None) -> None:
        self._base_dir = base_dir or DEFAULT_VERSION_CHECK_STATE_DIR.expanduser()

    def save(self, agent: str, version: str) -> Path:
        self._base_dir.mkdir(parents=True, exist_ok=True)
        path = self._base_dir / f"{_sanitize_agent_name(agent)}.yaml"
        with path.open("w", encoding="utf-8") as handle:
            payload = {
                _AGENT_KEY: agent,
                _VERSION_KEY: version,
                _CHECKED_AT_KEY: _now_iso(),
            }
            yaml.safe_dump(payload, handle, sort_keys=True)
        return path


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def _sanitize_agent_name(agent_name: str) -> str:
    return agent_name.replace("/", "_")



from __future__ import annotations

from typing import TYPE_CHECKING

from aicage._logging import get_logger
from aicage.docker.pull import run_pull
from aicage.docker.query import get_local_repo_digest_for_repo
from aicage.docker.remote_query import get_remote_repo_digest
from aicage.docker.types import ImageRefRepository, RegistryApiConfig, RemoteImageRef
from aicage.errors import CliError
from aicage.registry._logs import pull_log_path

if TYPE_CHECKING:
    from aicage.config.global_config import GlobalConfig


def refresh_base_digest(
    base_image_ref: str,
    base_repository: str,
    global_cfg: GlobalConfig,
) -> str | None:
    logger = get_logger()
    local_digest = get_local_repo_digest_for_repo(base_image_ref, base_repository)
    remote_digest = get_remote_repo_digest(
        RemoteImageRef(
            image=ImageRefRepository(
                image_ref=base_image_ref,
                repository=global_cfg.image_base_repository,
            ),
            registry_api=RegistryApiConfig(
                registry_api_url=global_cfg.image_registry_api_url,
                registry_api_token_url=global_cfg.image_registry_api_token_url,
            ),
        )
    )
    if remote_digest is None or remote_digest == local_digest:
        return local_digest

    log_path = pull_log_path(base_image_ref)
    try:
        run_pull(base_image_ref, log_path)
    except CliError:
        if local_digest:
            logger.warning(
                "Base image pull failed; using local base image (logs: %s).", log_path
            )
            return local_digest
        raise

    return get_local_repo_digest_for_repo(base_image_ref, base_repository)



from __future__ import annotations

from dataclasses import dataclass

from aicage.config.runtime_config import RunConfig
from aicage.docker.query import get_local_repo_digest
from aicage.docker.remote_query import get_remote_repo_digest
from aicage.docker.types import ImageRefRepository, RegistryApiConfig, RemoteImageRef


@dataclass(frozen=True)
class _PullDecision:
    should_pull: bool


def decide_pull(run_config: RunConfig) -> _PullDecision:
    # Local digests include registry prefix; registry API uses repository only.
    local_repository = f"{run_config.global_cfg.image_registry}/{run_config.global_cfg.image_repository}"
    remote_repository = run_config.global_cfg.image_repository
    local_digest = get_local_repo_digest(
        ImageRefRepository(
            image_ref=run_config.image_ref,
            repository=local_repository,
        )
    )
    if local_digest is None:
        return _PullDecision(should_pull=True)

    remote_digest = get_remote_repo_digest(
        RemoteImageRef(
            image=ImageRefRepository(
                image_ref=run_config.image_ref,
                repository=remote_repository,
            ),
            registry_api=RegistryApiConfig(
                registry_api_url=run_config.global_cfg.image_registry_api_url,
                registry_api_token_url=run_config.global_cfg.image_registry_api_token_url,
            ),
        )
    )
    if remote_digest is None:
        return _PullDecision(should_pull=False)

    return _PullDecision(should_pull=local_digest != remote_digest)



from __future__ import annotations

import os
from datetime import datetime, timezone
from pathlib import Path

_DEFAULT_LOG_DIR = "~/.aicage/logs/pull"


def pull_log_path(image_ref: str) -> Path:
    log_dir = Path(os.path.expanduser(_DEFAULT_LOG_DIR))
    return log_dir / f"{_sanitize(image_ref)}-{_timestamp()}.log"


def _sanitize(value: str) -> str:
    return value.replace("/", "_").replace(":", "_")


def _timestamp() -> str:
    return datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")



import os
from dataclasses import dataclass
from pathlib import Path

from aicage.errors import CliError
from aicage.registry.images_metadata.models import AgentMetadata, ImagesMetadata


@dataclass
class AgentConfig:
    agent_path: str
    agent_config_host: Path


def resolve_agent_config(agent: str, images_metadata: ImagesMetadata) -> AgentConfig:
    agent_path = _read_agent_path(agent, images_metadata)
    agent_config_host = Path(os.path.expanduser(agent_path)).resolve()
    agent_config_host.mkdir(parents=True, exist_ok=True)
    return AgentConfig(agent_path=agent_path, agent_config_host=agent_config_host)


def _read_agent_path(agent: str, images_metadata: ImagesMetadata) -> str:
    agent_metadata = _require_agent_metadata(agent, images_metadata)
    return agent_metadata.agent_path


def _require_agent_metadata(agent: str, images_metadata: ImagesMetadata) -> AgentMetadata:
    agent_metadata = images_metadata.agents.get(agent)
    if not agent_metadata:
        raise CliError(f"Agent '{agent}' is missing from images metadata.")
    return agent_metadata



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml

from aicage._logging import get_logger
from aicage.errors import CliError
from aicage.paths import DEFAULT_CUSTOM_EXTENDED_IMAGES_DIR, EXTENDED_IMAGE_DEFINITION_FILENAME

_AGENT_KEY: str = "agent"
_BASE_KEY: str = "base"
_EXTENSIONS_KEY: str = "extensions"
_IMAGE_REF_KEY: str = "image_ref"


@dataclass(frozen=True)
class ExtendedImageConfig:
    name: str
    agent: str
    base: str
    extensions: list[str]
    image_ref: str
    path: Path


def load_extended_images(available_extensions: set[str]) -> dict[str, ExtendedImageConfig]:
    images_dir = DEFAULT_CUSTOM_EXTENDED_IMAGES_DIR.expanduser()
    if not images_dir.is_dir():
        return {}
    configs: dict[str, ExtendedImageConfig] = {}
    logger = get_logger()
    for entry in sorted(images_dir.iterdir()):
        if not entry.is_dir():
            continue
        config_path = entry / EXTENDED_IMAGE_DEFINITION_FILENAME
        if not config_path.is_file():
            raise CliError(
                f"Extended image '{entry.name}' is missing {EXTENDED_IMAGE_DEFINITION_FILENAME}."
            )
        mapping = _load_yaml(config_path)
        _expect_keys(
            mapping,
            required={_AGENT_KEY, _BASE_KEY, _EXTENSIONS_KEY, _IMAGE_REF_KEY},
            optional=set(),
            context=f"extended image config at {config_path}",
        )
        extensions = _read_str_list(mapping.get(_EXTENSIONS_KEY), _EXTENSIONS_KEY)
        missing = [ext for ext in extensions if ext not in available_extensions]
        if missing:
            logger.warning(
                "Skipping extended image %s; missing extensions: %s",
                entry.name,
                ", ".join(sorted(missing)),
            )
            continue
        configs[entry.name] = ExtendedImageConfig(
            name=entry.name,
            agent=_expect_string(mapping.get(_AGENT_KEY), _AGENT_KEY),
            base=_expect_string(mapping.get(_BASE_KEY), _BASE_KEY),
            extensions=extensions,
            image_ref=_expect_string(mapping.get(_IMAGE_REF_KEY), _IMAGE_REF_KEY),
            path=config_path,
        )
    return configs


def write_extended_image_config(config: ExtendedImageConfig) -> None:
    config.path.parent.mkdir(parents=True, exist_ok=True)
    payload = {
        _AGENT_KEY: config.agent,
        _BASE_KEY: config.base,
        _EXTENSIONS_KEY: list(config.extensions),
        _IMAGE_REF_KEY: config.image_ref,
    }
    config.path.write_text(yaml.safe_dump(payload, sort_keys=True), encoding="utf-8")


def extended_image_config_path(name: str) -> Path:
    return (
        DEFAULT_CUSTOM_EXTENDED_IMAGES_DIR.expanduser()
        / name
        / EXTENDED_IMAGE_DEFINITION_FILENAME
    )


def _load_yaml(path: Path) -> dict[str, Any]:
    try:
        payload = path.read_text(encoding="utf-8")
        data = yaml.safe_load(payload) or {}
    except (OSError, yaml.YAMLError) as exc:
        raise CliError(f"Failed to read extended image config from {path}: {exc}") from exc
    if not isinstance(data, dict):
        raise CliError(f"Extended image config at {path} must be a mapping.")
    return data


def _expect_string(value: Any, context: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise CliError(f"{context} must be a non-empty string.")
    return value


def _read_str_list(value: Any, context: str) -> list[str]:
    if not isinstance(value, list):
        raise CliError(f"{context} must be a list.")
    items: list[str] = []
    for item in value:
        if not isinstance(item, str) or not item.strip():
            raise CliError(f"{context} must contain non-empty strings.")
        items.append(item)
    return items


def _expect_keys(
    mapping: dict[str, Any],
    required: set[str],
    optional: set[str],
    context: str,
) -> None:
    missing = sorted(required - set(mapping))
    if missing:
        raise CliError(f"{context} missing required keys: {', '.join(missing)}.")
    unknown = sorted(set(mapping) - required - optional)
    if unknown:
        raise CliError(f"{context} contains unsupported keys: {', '.join(unknown)}.")



from dataclasses import dataclass

from aicage.errors import CliError

from ._tty import ensure_tty_for_prompt


@dataclass(frozen=True)
class ExtensionOption:
    name: str
    description: str


def prompt_for_extensions(options: list[ExtensionOption]) -> list[str]:
    if not options:
        return []
    ensure_tty_for_prompt()
    print("Select extensions to add (comma-separated numbers or names, empty for none):")
    for idx, option in enumerate(options, start=1):
        print(f"  {idx}) {option.name}: {option.description}")
    response = input("Enter selection: ").strip()
    if not response:
        return []
    requested = [item.strip() for item in response.split(",") if item.strip()]
    selection: list[str] = []
    seen: set[str] = set()
    for item in requested:
        extension_id = _resolve_extension_choice(item, options)
        if extension_id in seen:
            raise CliError(f"Duplicate extension '{extension_id}' selected.")
        seen.add(extension_id)
        selection.append(extension_id)
    return selection


def _resolve_extension_choice(response: str, options: list[ExtensionOption]) -> str:
    if response.isdigit():
        idx = int(response)
        if idx < 1 or idx > len(options):
            raise CliError(f"Invalid selection '{response}'. Pick a number between 1 and {len(options)}.")
        return options[idx - 1].name
    for option in options:
        if option.name == response:
            return option.name
    valid = ", ".join(option.name for option in options)
    raise CliError(f"Invalid extension '{response}'. Valid options: {valid}")



from dataclasses import dataclass
from pathlib import Path

from aicage.registry.images_metadata.loader import load_images_metadata
from aicage.registry.images_metadata.models import ImagesMetadata

from .config_store import SettingsStore
from .global_config import GlobalConfig
from .project_config import ProjectConfig


@dataclass
class ConfigContext:
    store: SettingsStore
    project_cfg: ProjectConfig
    global_cfg: GlobalConfig
    images_metadata: ImagesMetadata

    def image_repository_ref(self) -> str:
        return f"{self.global_cfg.image_registry}/{self.global_cfg.image_repository}"


def _build_config_context() -> ConfigContext:
    store = SettingsStore()
    project_path = Path.cwd().resolve()
    global_cfg = store.load_global()
    images_metadata = load_images_metadata(global_cfg.local_image_repository)
    project_cfg = store.load_project(project_path)
    return ConfigContext(
        store=store,
        project_cfg=project_cfg,
        global_cfg=global_cfg,
        images_metadata=images_metadata,
    )



from pathlib import Path

from aicage.registry._extended_images import (
    ExtendedImageConfig,
    extended_image_config_path,
    write_extended_image_config,
)
from aicage.runtime.prompts import ExtensionOption, prompt_for_extensions, prompt_for_image_ref

from ..models import ImageSelection
from .context import ExtensionSelectionContext
from .refs import base_image_ref


def handle_extension_selection(selection: ExtensionSelectionContext) -> ImageSelection:
    agent_cfg = selection.agent_cfg
    agent_cfg.base = selection.base
    extension_options = [
        ExtensionOption(
            name=ext.extension_id,
            description=f"{ext.name}: {ext.description}",
        )
        for ext in sorted(selection.extensions.values(), key=lambda item: item.extension_id)
    ]
    selected_extensions = prompt_for_extensions(extension_options) if extension_options else []
    if selected_extensions:
        image_ref = prompt_for_image_ref(
            _default_extended_image_ref(selection.agent, selection.base, selected_extensions)
        )
        agent_cfg.extensions = list(selected_extensions)
        agent_cfg.image_ref = image_ref
        write_extended_image_config(
            ExtendedImageConfig(
                name=_extended_image_name(image_ref),
                agent=selection.agent,
                base=selection.base,
                extensions=list(selected_extensions),
                image_ref=image_ref,
                path=extended_image_config_path(_extended_image_name(image_ref)),
            )
        )
    else:
        agent_cfg.extensions = []
        agent_cfg.image_ref = base_image_ref(
            selection.agent_metadata,
            selection.agent,
            selection.base,
            selection.context,
        )
    selection.context.store.save_project(
        Path(selection.context.project_cfg.path),
        selection.context.project_cfg,
    )
    return ImageSelection(
        image_ref=agent_cfg.image_ref or "",
        base=selection.base,
        extensions=list(agent_cfg.extensions),
        base_image_ref=base_image_ref(
            selection.agent_metadata,
            selection.agent,
            selection.base,
            selection.context,
        ),
    )


def _default_extended_image_ref(agent: str, base: str, extensions: list[str]) -> str:
    tag = "-".join([agent, base, *extensions]).lower().replace("/", "-")
    return f"aicage-extended:{tag}"


def _extended_image_name(image_ref: str) -> str:
    _, _, tag = image_ref.rpartition(":")
    return tag or image_ref



import sys
from collections.abc import Sequence

from aicage._logging import get_logger
from aicage.cli._parse import parse_cli
from aicage.cli._print_config import print_project_config
from aicage.cli_types import ParsedArgs
from aicage.config import ConfigError
from aicage.config.runtime_config import RunConfig, load_run_config
from aicage.docker.run import print_run_command, run_container
from aicage.errors import CliError
from aicage.registry.image_pull import pull_image
from aicage.registry.local_build.ensure_extended_image import ensure_extended_image
from aicage.registry.local_build.ensure_local_image import ensure_local_image
from aicage.runtime.run_args import DockerRunArgs
from aicage.runtime.run_plan import build_run_args


def main(argv: Sequence[str] | None = None) -> int:
    parsed_argv: Sequence[str] = argv if argv is not None else sys.argv[1:]
    logger = get_logger()
    try:
        parsed: ParsedArgs = parse_cli(parsed_argv)
        if parsed.config_action == "print":
            print_project_config()
            return 0
        run_config: RunConfig = load_run_config(parsed.agent, parsed)
        logger.info("Resolved run config for agent %s", run_config.agent)
        agent_metadata = run_config.images_metadata.agents[run_config.agent]
        if run_config.extensions:
            ensure_extended_image(run_config)
        elif agent_metadata.local_definition_dir is None:
            pull_image(run_config)
        else:
            ensure_local_image(run_config)
        run_args: DockerRunArgs = build_run_args(config=run_config, parsed=parsed)

        if parsed.dry_run:
            print_run_command(run_args)
            logger.info("Dry-run docker command printed.")
            return 0

        run_container(run_args)
        return 0
    except KeyboardInterrupt:
        print()
        logger.warning("Interrupted by user.")
        return 130
    except (CliError, ConfigError) as exc:
        print(f"[aicage] {exc}", file=sys.stderr)
        logger.error("CLI error: %s", exc)
        return 1



from __future__ import annotations

import os
import subprocess
from dataclasses import dataclass
from logging import Logger
from pathlib import Path

from aicage._logging import get_logger
from aicage.config.global_config import GlobalConfig
from aicage.docker.pull import run_pull
from aicage.docker.query import get_local_repo_digest
from aicage.docker.remote_query import get_remote_repo_digest
from aicage.docker.run import run_builder_version_check
from aicage.docker.types import ImageRefRepository, RegistryApiConfig, RemoteImageRef
from aicage.errors import CliError
from aicage.registry._logs import pull_log_path
from aicage.registry.images_metadata.models import AgentMetadata

from .store import VersionCheckStore


class AgentVersionChecker:
    def __init__(self, global_cfg: GlobalConfig, store: VersionCheckStore | None = None) -> None:
        self._global_cfg = global_cfg
        self._store = store or VersionCheckStore()

    def get_version(
        self,
        agent_name: str,
        _agent_metadata: AgentMetadata,
        definition_dir: Path,
    ) -> str:
        logger = get_logger()
        script_path = definition_dir / "version.sh"
        if not script_path.is_file():
            raise CliError(f"Agent '{agent_name}' is missing version.sh at {script_path}.")

        errors: list[str] = []
        host_result = _run_host(script_path)
        if host_result.success:
            logger.info("Version check succeeded on host for %s", agent_name)
            self._store.save(agent_name, host_result.output)
            return host_result.output

        logger.warning(
            "Version check failed on host for %s: %s",
            agent_name,
            host_result.error,
        )
        errors.append(host_result.error)

        builder_result = _run_builder(
            image_ref=self._global_cfg.version_check_image,
            definition_dir=definition_dir,
            global_cfg=self._global_cfg,
        )
        if builder_result.success:
            logger.info("Version check succeeded in builder for %s", agent_name)
            self._store.save(agent_name, builder_result.output)
            return builder_result.output

        logger.warning(
            "Version check failed in builder for %s: %s",
            agent_name,
            builder_result.error,
        )
        errors.append(builder_result.error)
        logger.error("Version check failed for %s: %s", agent_name, "; ".join(errors))
        raise CliError("; ".join(errors))


@dataclass(frozen=True)
class _CommandResult:
    success: bool
    output: str
    error: str


def _run_builder(image_ref: str, definition_dir: Path, global_cfg: GlobalConfig) -> _CommandResult:
    _ensure_version_check_image(image_ref, global_cfg, get_logger())
    process = run_builder_version_check(image_ref, definition_dir)
    return _from_process(process, "builder image")


def _run_host(script_path: Path) -> _CommandResult:
    if not os.access(script_path, os.X_OK):
        get_logger().warning(
            "version.sh at %s is not executable; running with /bin/bash.",
            script_path,
        )
    return _run_command(["/bin/bash", str(script_path)], "host")


def _run_command(command: list[str], context: str) -> _CommandResult:
    process = subprocess.run(command, check=False, capture_output=True, text=True)
    return _from_process(process, context)


def _from_process(process: subprocess.CompletedProcess[str], context: str) -> _CommandResult:
    output = process.stdout.strip() if process.stdout else ""
    if process.returncode == 0 and output:
        return _CommandResult(success=True, output=output, error="")

    stderr = process.stderr.strip() if process.stderr else ""
    error = stderr or output or f"Version check failed in {context}."
    return _CommandResult(success=False, output=output, error=error)


def _ensure_version_check_image(image_ref: str, global_cfg: GlobalConfig, logger: Logger) -> None:
    local_image, remote_image = _version_check_images(image_ref, global_cfg)
    local_digest = get_local_repo_digest(local_image)
    if local_digest is None:
        _pull_version_check_image(image_ref, logger)
        return

    remote_digest = get_remote_repo_digest(remote_image)
    if remote_digest is None or remote_digest == local_digest:
        return

    _pull_version_check_image(image_ref, logger)


def _pull_version_check_image(image_ref: str, logger: Logger) -> None:
    log_path = pull_log_path(image_ref)
    try:
        run_pull(image_ref, log_path)
    except CliError:
        logger.warning("Version check image pull failed; using local image (logs: %s).", log_path)


def _version_check_images(
    image_ref: str, global_cfg: GlobalConfig
) -> tuple[ImageRefRepository, RemoteImageRef]:
    registry, repository = _split_image_ref(image_ref, global_cfg.image_registry)
    local_repository = f"{registry}/{repository}" if registry else repository
    local_image = ImageRefRepository(image_ref=image_ref, repository=local_repository)
    remote_image = RemoteImageRef(
        image=ImageRefRepository(image_ref=image_ref, repository=repository),
        registry_api=RegistryApiConfig(
            registry_api_url=global_cfg.image_registry_api_url,
            registry_api_token_url=global_cfg.image_registry_api_token_url,
        ),
    )
    return local_image, remote_image


def _split_image_ref(image_ref: str, default_registry: str) -> tuple[str, str]:
    name = _strip_reference(image_ref)
    parts = name.split("/", 1)
    if len(parts) == 1:
        return default_registry, name
    registry, remainder = parts
    if "." in registry or ":" in registry or registry == "localhost":
        return registry, remainder
    return default_registry, name


def _strip_reference(image_ref: str) -> str:
    if "@" in image_ref:
        return image_ref.split("@", 1)[0]
    last_colon = image_ref.rfind(":")
    if last_colon > image_ref.rfind("/"):
        return image_ref[:last_colon]
    return image_ref



from pathlib import Path

from aicage.config.project_config import AgentConfig
from aicage.runtime.prompts import prompt_persist_docker_socket
from aicage.runtime.run_args import MountSpec

_DOCKER_SOCKET_PATH = Path("/run/docker.sock")


def resolve_docker_socket_mount(
    agent_cfg: AgentConfig,
    cli_docker_socket: bool,
) -> list[MountSpec]:
    mounts_cfg = agent_cfg.mounts
    docker_socket_enabled = cli_docker_socket or bool(mounts_cfg.docker)
    if not docker_socket_enabled:
        return []

    mounts = [
        MountSpec(
            host_path=_DOCKER_SOCKET_PATH,
            container_path=_DOCKER_SOCKET_PATH,
        )
    ]

    if cli_docker_socket and mounts_cfg.docker is None:
        if prompt_persist_docker_socket():
            mounts_cfg.docker = True

    return mounts



from pathlib import Path

from ._exec import capture_stdout


def is_commit_signing_enabled(repo_path: Path) -> bool:
    stdout = capture_stdout(["git", "config", "commit.gpgsign"], cwd=repo_path)
    if not stdout:
        return False
    value = stdout.strip().lower()
    return value in {"true", "1", "yes", "on"}


def resolve_signing_format(repo_path: Path) -> str | None:
    stdout = capture_stdout(["git", "config", "gpg.format"], cwd=repo_path)
    if not stdout:
        return None
    value = stdout.strip().lower()
    return value or None



from __future__ import annotations

from pathlib import Path

CONFIG_FILENAME: str = "config.yaml"
IMAGES_METADATA_FILENAME: str = "images-metadata.yaml"
_AGENT_DEFINITION_FILENAME: str = "agent.yaml"

DEFAULT_CUSTOM_AGENTS_DIR: Path = Path("~/.aicage/custom/agents")
CUSTOM_AGENT_DEFINITION_FILES: tuple[str, str] = (
    _AGENT_DEFINITION_FILENAME,
    "agent.yml",
)

DEFAULT_CUSTOM_EXTENSIONS_DIR: Path = Path("~/.aicage/custom/extensions")
CUSTOM_EXTENSION_DEFINITION_FILES: tuple[str, str] = (
    "extension.yaml",
    "extension.yml",
)
DEFAULT_CUSTOM_EXTENDED_IMAGES_DIR: Path = Path("~/.aicage/custom/image-extended")
EXTENDED_IMAGE_DEFINITION_FILENAME: str = "image-extended.yaml"

DEFAULT_VERSION_CHECK_STATE_DIR: Path = Path("~/.aicage/state/version-check")
DEFAULT_LOCAL_BUILD_STATE_DIR: Path = Path("~/.aicage/state/local-build")
DEFAULT_LOCAL_EXTENDED_STATE_DIR: Path = Path("~/.aicage/state/local-extended")



from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml

from aicage.errors import CliError
from aicage.paths import CUSTOM_AGENT_DEFINITION_FILES, DEFAULT_CUSTOM_AGENTS_DIR
from aicage.registry._image_refs import local_image_ref
from aicage.registry.images_metadata.models import (
    AGENT_FULL_NAME_KEY,
    AGENT_HOMEPAGE_KEY,
    AGENT_PATH_KEY,
    BASE_DISTRO_EXCLUDE_KEY,
    BASE_EXCLUDE_KEY,
    AgentMetadata,
    ImagesMetadata,
)

from ._validation import ensure_required_files, expect_string, maybe_str_list, validate_agent_mapping


def load_custom_agents(
    images_metadata: ImagesMetadata,
    local_image_repository: str,
) -> dict[str, AgentMetadata]:
    agents_dir = DEFAULT_CUSTOM_AGENTS_DIR.expanduser()
    if not agents_dir.is_dir():
        return {}

    custom_agents: dict[str, AgentMetadata] = {}
    for entry in sorted(agents_dir.iterdir()):
        if not entry.is_dir():
            continue
        agent_name = entry.name
        agent_path = _find_agent_definition(entry)
        agent_mapping = _load_yaml(agent_path)
        ensure_required_files(agent_name, entry)
        custom_agents[agent_name] = _build_custom_agent(
            agent_name=agent_name,
            agent_mapping=agent_mapping,
            images_metadata=images_metadata,
            local_image_repository=local_image_repository,
        )
    return custom_agents


def _find_agent_definition(agent_dir: Path) -> Path:
    for filename in CUSTOM_AGENT_DEFINITION_FILES:
        candidate = agent_dir / filename
        if candidate.is_file():
            return candidate
    expected = ", ".join(CUSTOM_AGENT_DEFINITION_FILES)
    raise CliError(f"Custom agent '{agent_dir.name}' is missing {expected}.")


def _load_yaml(path: Path) -> dict[str, Any]:
    try:
        payload = path.read_text(encoding="utf-8")
        data = yaml.safe_load(payload) or {}
    except (OSError, yaml.YAMLError) as exc:
        raise CliError(f"Failed to read custom agent metadata from {path}: {exc}") from exc
    if not isinstance(data, dict):
        raise CliError(f"Custom agent metadata at {path} must be a mapping.")
    return data


def _build_custom_agent(
    agent_name: str,
    agent_mapping: dict[str, Any],
    images_metadata: ImagesMetadata,
    local_image_repository: str,
) -> AgentMetadata:
    normalized_mapping = validate_agent_mapping(agent_mapping)
    base_exclude = maybe_str_list(normalized_mapping.get(BASE_EXCLUDE_KEY), BASE_EXCLUDE_KEY)
    base_distro_exclude = maybe_str_list(
        normalized_mapping.get(BASE_DISTRO_EXCLUDE_KEY), BASE_DISTRO_EXCLUDE_KEY
    )
    valid_bases = _build_valid_bases(
        agent_name=agent_name,
        images_metadata=images_metadata,
        base_exclude=base_exclude,
        base_distro_exclude=base_distro_exclude,
        local_image_repository=local_image_repository,
    )
    return AgentMetadata(
        agent_path=expect_string(normalized_mapping.get(AGENT_PATH_KEY), AGENT_PATH_KEY),
        agent_full_name=expect_string(normalized_mapping.get(AGENT_FULL_NAME_KEY), AGENT_FULL_NAME_KEY),
        agent_homepage=expect_string(normalized_mapping.get(AGENT_HOMEPAGE_KEY), AGENT_HOMEPAGE_KEY),
        valid_bases=valid_bases,
        base_exclude=base_exclude,
        base_distro_exclude=base_distro_exclude,
        local_definition_dir=DEFAULT_CUSTOM_AGENTS_DIR.expanduser() / agent_name,
    )


def _build_valid_bases(
    agent_name: str,
    images_metadata: ImagesMetadata,
    base_exclude: list[str] | None,
    base_distro_exclude: list[str] | None,
    local_image_repository: str,
) -> dict[str, str]:
    valid_bases: dict[str, str] = {}
    base_exclude_set = _normalize_exclude(base_exclude)
    base_distro_exclude_set = _normalize_exclude(base_distro_exclude)
    for base_name in sorted(images_metadata.bases):
        base_metadata = images_metadata.bases[base_name]
        if _is_base_excluded(
            base_name,
            base_metadata.base_image_distro,
            base_exclude_set,
            base_distro_exclude_set,
        ):
            continue
        valid_bases[base_name] = local_image_ref(local_image_repository, agent_name, base_name)
    return valid_bases


def _is_base_excluded(
    base_name: str,
    base_distro: str,
    base_exclude: set[str],
    base_distro_exclude: set[str],
) -> bool:
    base_name_lc = base_name.lower()
    if base_name_lc in base_exclude:
        return True
    if base_distro.lower() in base_distro_exclude:
        return True
    return False


def _normalize_exclude(values: list[str] | None) -> set[str]:
    if not values:
        return set()
    return {value.lower() for value in values}



from __future__ import annotations

import json
from functools import lru_cache
from pathlib import Path
from typing import Any

from aicage.config.resources import find_packaged_path
from aicage.errors import CliError
from aicage.registry.images_metadata.models import BUILD_LOCAL_KEY

_AGENT_SCHEMA_PATH = "validation/agent.schema.json"
_CUSTOM_AGENT_CONTEXT = "custom agent metadata"


def validate_agent_mapping(mapping: dict[str, Any]) -> dict[str, Any]:
    context = _CUSTOM_AGENT_CONTEXT
    if not isinstance(mapping, dict):
        raise CliError(f"{context} must be a mapping.")

    schema = _load_schema()
    properties = schema.get("properties", {})
    required = set(schema.get("required", []))
    additional = schema.get("additionalProperties", True)

    missing = sorted(required - set(mapping))
    if missing:
        raise CliError(f"{context} missing required keys: {', '.join(missing)}.")

    if additional is False:
        unknown = sorted(set(mapping) - set(properties))
        if unknown:
            raise CliError(f"{context} contains unsupported keys: {', '.join(unknown)}.")

    normalized = dict(mapping)
    normalized.setdefault(BUILD_LOCAL_KEY, True)

    for key, value in normalized.items():
        schema_entry = properties.get(key)
        if schema_entry is None:
            continue
        _validate_value(value, schema_entry, f"{context}.{key}")

    return normalized


def ensure_required_files(agent_name: str, agent_dir: Path) -> None:
    missing = [name for name in ("install.sh", "version.sh") if not (agent_dir / name).is_file()]
    if missing:
        raise CliError(f"Custom agent '{agent_name}' is missing {', '.join(missing)}.")


@lru_cache(maxsize=1)
def _load_schema() -> dict[str, Any]:
    path = find_packaged_path(_AGENT_SCHEMA_PATH)
    payload = path.read_text(encoding="utf-8")
    return json.loads(payload)


def _validate_value(value: Any, schema_entry: dict[str, Any], context: str) -> None:
    schema_type = schema_entry.get("type")
    if schema_type == "string":
        expect_string(value, context)
        return
    if schema_type == "boolean":
        _expect_bool(value, context)
        return
    if schema_type == "array":
        _expect_str_list(value, context, schema_entry)
        return
    raise CliError(f"{context} has unsupported schema type '{schema_type}'.")


def _expect_str_list(value: Any, context: str, schema_entry: dict[str, Any]) -> None:
    if not isinstance(value, list):
        raise CliError(f"{context} must be a list.")
    item_schema = schema_entry.get("items", {})
    item_type = item_schema.get("type")
    if item_type != "string":
        raise CliError(f"{context} items must be strings.")
    for item in value:
        expect_string(item, context)


def expect_string(value: Any, context: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise CliError(f"{context} must be a non-empty string.")
    return value


def _expect_bool(value: Any, context: str) -> bool:
    if not isinstance(value, bool):
        raise CliError(f"{context} must be a boolean.")
    return value


def maybe_str_list(value: Any, context: str) -> list[str] | None:
    if value is None:
        return None
    if not isinstance(value, list):
        raise CliError(f"{context} must be a list.")
    items: list[str] = []
    for item in value:
        if not isinstance(item, str) or not item.strip():
            raise CliError(f"{context} must contain non-empty strings.")
        items.append(item)
    return items



from dataclasses import dataclass

from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.registry.extensions import ExtensionMetadata
from aicage.registry.images_metadata.models import AgentMetadata


@dataclass(frozen=True)
class ExtensionSelectionContext:
    agent: str
    base: str
    agent_cfg: AgentConfig
    agent_metadata: AgentMetadata
    extensions: dict[str, ExtensionMetadata]
    context: ConfigContext



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import yaml

from aicage.paths import DEFAULT_LOCAL_EXTENDED_STATE_DIR

from ._store import sanitize

_AGENT_KEY: str = "agent"
_BASE_KEY: str = "base"
_IMAGE_REF_KEY: str = "image_ref"
_EXTENSIONS_KEY: str = "extensions"
_EXTENSION_HASH_KEY: str = "extension_hash"
_BASE_IMAGE_KEY: str = "base_image"
_BUILT_AT_KEY: str = "built_at"


@dataclass(frozen=True)
class ExtendedBuildRecord:
    agent: str
    base: str
    image_ref: str
    extensions: list[str]
    extension_hash: str
    base_image: str
    built_at: str


class ExtendedBuildStore:
    def __init__(self, base_dir: Path | None = None) -> None:
        self._base_dir = base_dir or DEFAULT_LOCAL_EXTENDED_STATE_DIR.expanduser()

    def load(self, image_ref: str) -> ExtendedBuildRecord | None:
        path = self._path(image_ref)
        if not path.is_file():
            return None
        payload = yaml.safe_load(path.read_text(encoding="utf-8")) or {}
        if not isinstance(payload, dict):
            return None
        return ExtendedBuildRecord(
            agent=str(payload.get(_AGENT_KEY, "")),
            base=str(payload.get(_BASE_KEY, "")),
            image_ref=str(payload.get(_IMAGE_REF_KEY, "")),
            extensions=_read_str_list(payload.get(_EXTENSIONS_KEY)),
            extension_hash=str(payload.get(_EXTENSION_HASH_KEY, "")),
            base_image=str(payload.get(_BASE_IMAGE_KEY, "")),
            built_at=str(payload.get(_BUILT_AT_KEY, "")),
        )

    def save(self, record: ExtendedBuildRecord) -> Path:
        self._base_dir.mkdir(parents=True, exist_ok=True)
        path = self._path(record.image_ref)
        payload = {
            _AGENT_KEY: record.agent,
            _BASE_KEY: record.base,
            _IMAGE_REF_KEY: record.image_ref,
            _EXTENSIONS_KEY: list(record.extensions),
            _EXTENSION_HASH_KEY: record.extension_hash,
            _BASE_IMAGE_KEY: record.base_image,
            _BUILT_AT_KEY: record.built_at,
        }
        path.write_text(yaml.safe_dump(payload, sort_keys=True), encoding="utf-8")
        return path

    def _path(self, image_ref: str) -> Path:
        filename = f"{sanitize(image_ref)}.yaml"
        return self._base_dir / filename


def _read_str_list(value: object) -> list[str]:
    if not isinstance(value, list):
        return []
    return [item for item in value if isinstance(item, str) and item]



from .config_store import SettingsStore as SettingsStore
from .errors import ConfigError as ConfigError
from .global_config import GlobalConfig as GlobalConfig
from .project_config import ProjectConfig as ProjectConfig



from pathlib import Path

from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.errors import CliError
from aicage.registry._extended_images import load_extended_images
from aicage.registry.extensions import ExtensionMetadata
from aicage.registry.images_metadata.models import AgentMetadata
from aicage.runtime.prompts import ExtendedImageOption

from ..models import ImageSelection
from .refs import base_image_ref


def load_extended_image_options(
    agent: str,
    agent_metadata: AgentMetadata,
    extensions: dict[str, ExtensionMetadata],
) -> list[ExtendedImageOption]:
    configs = load_extended_images(set(extensions))
    options: list[ExtendedImageOption] = []
    for config in configs.values():
        if config.agent != agent:
            continue
        if config.base not in agent_metadata.valid_bases:
            continue
        options.append(
            ExtendedImageOption(
                name=config.name,
                base=config.base,
                description="Custom extended image",
                extensions=list(config.extensions),
                image_ref=config.image_ref,
            )
        )
    return options


def resolve_extended_image(
    name: str,
    options: list[ExtendedImageOption],
) -> ExtendedImageOption:
    for option in options:
        if option.name == name:
            return option
    raise CliError(f"Unknown extended image '{name}'.")


def apply_extended_selection(
    agent: str,
    agent_cfg: AgentConfig,
    selected: ExtendedImageOption,
    agent_metadata: AgentMetadata,
    context: ConfigContext,
) -> ImageSelection:
    agent_cfg.base = selected.base
    agent_cfg.extensions = list(selected.extensions)
    agent_cfg.image_ref = selected.image_ref
    context.store.save_project(Path(context.project_cfg.path), context.project_cfg)
    return ImageSelection(
        image_ref=selected.image_ref,
        base=selected.base,
        extensions=list(selected.extensions),
        base_image_ref=base_image_ref(agent_metadata, agent, selected.base, context),
    )



from .base import BaseOption as BaseOption
from .base import BaseSelectionRequest as BaseSelectionRequest
from .base import available_bases as available_bases
from .base import base_options as base_options
from .base import prompt_for_base as prompt_for_base
from .confirm import prompt_mount_git_config as prompt_mount_git_config
from .confirm import prompt_mount_gpg_keys as prompt_mount_gpg_keys
from .confirm import prompt_mount_ssh_keys as prompt_mount_ssh_keys
from .confirm import prompt_persist_docker_args as prompt_persist_docker_args
from .confirm import prompt_persist_docker_socket as prompt_persist_docker_socket
from .confirm import prompt_persist_entrypoint as prompt_persist_entrypoint
from .confirm import prompt_yes_no as prompt_yes_no
from .extensions import ExtensionOption as ExtensionOption
from .extensions import prompt_for_extensions as prompt_for_extensions
from .image_choice import ExtendedImageOption as ExtendedImageOption
from .image_choice import ImageChoice as ImageChoice
from .image_choice import ImageChoiceRequest as ImageChoiceRequest
from .image_choice import prompt_for_image_choice as prompt_for_image_choice
from .image_ref import prompt_for_image_ref as prompt_for_image_ref
from .missing_extensions import prompt_for_missing_extensions as prompt_for_missing_extensions



from __future__ import annotations

from aicage.registry.custom_agent import load_custom_agents
from aicage.registry.images_metadata.models import ImagesMetadata


def discover_agents(
    images_metadata: ImagesMetadata,
    local_image_repository: str,
) -> ImagesMetadata:
    custom_agents = load_custom_agents(images_metadata, local_image_repository)
    if not custom_agents:
        return images_metadata

    merged_agents = dict(images_metadata.agents)
    merged_agents.update(custom_agents)
    return ImagesMetadata(
        aicage_image=images_metadata.aicage_image,
        aicage_image_base=images_metadata.aicage_image_base,
        bases=images_metadata.bases,
        agents=merged_agents,
    )



from __future__ import annotations

from datetime import datetime, timezone

from aicage._logging import get_logger
from aicage.config.runtime_config import RunConfig
from aicage.docker.query import get_local_rootfs_layers, local_image_exists

from ._store import BuildRecord


def should_build(
    run_config: RunConfig,
    record: BuildRecord | None,
    base_image_ref: str,
) -> bool:
    if not local_image_exists(run_config.image_ref):
        return True
    if record is None:
        return True
    if record.agent_version != run_config.agent_version:
        return True
    base_layer_missing = _base_layer_missing(base_image_ref, run_config.image_ref)
    if base_layer_missing is None:
        logger = get_logger()
        logger.warning(
            "Skipping base image layer validation for %s; missing local layer data.",
            run_config.image_ref,
        )
        return False
    if base_layer_missing:
        return True
    return False


def base_repository(run_config: RunConfig) -> str:
    return f"{run_config.global_cfg.image_registry}/{run_config.global_cfg.image_base_repository}"


def base_image_ref(run_config: RunConfig) -> str:
    repository = base_repository(run_config)
    return f"{repository}:{run_config.base}"


def now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def _base_layer_missing(base_image_ref: str, final_image_ref: str) -> bool | None:
    base_layers = get_local_rootfs_layers(base_image_ref)
    if base_layers is None:
        return None
    final_layers = get_local_rootfs_layers(final_image_ref)
    if final_layers is None:
        return None
    return base_layers[-1] not in final_layers



from __future__ import annotations

from pathlib import Path

from aicage.config.context import ConfigContext
from aicage.config.project_config import AGENT_BASE_KEY, AgentConfig
from aicage.errors import CliError
from aicage.registry.extensions import ExtensionMetadata, load_extensions
from aicage.registry.images_metadata.models import AgentMetadata, ImagesMetadata
from aicage.runtime.prompts import (
    BaseSelectionRequest,
    ImageChoice,
    ImageChoiceRequest,
    prompt_for_base,
    prompt_for_image_choice,
)

from .extensions.context import ExtensionSelectionContext
from .extensions.extended_images import (
    apply_extended_selection,
    load_extended_image_options,
    resolve_extended_image,
)
from .extensions.handler import handle_extension_selection
from .extensions.missing_extensions import ensure_extensions_exist
from .extensions.refs import base_image_ref
from .models import ImageSelection


def select_agent_image(agent: str, context: ConfigContext) -> ImageSelection:
    agent_cfg = context.project_cfg.agents.setdefault(agent, AgentConfig())
    agent_metadata = _require_agent_metadata(agent, context.images_metadata)
    base = agent_cfg.base or context.global_cfg.agents.get(agent, {}).get(AGENT_BASE_KEY)
    extensions = load_extensions()

    if agent_cfg.image_ref:
        if base is None:
            return _fresh_selection(agent, context, agent_metadata, extensions)
        _validate_base(agent, base, agent_metadata)
        if agent_cfg.extensions:
            reset = ensure_extensions_exist(
                agent=agent,
                project_config_path=context.store.project_config_path(Path(context.project_cfg.path)),
                agent_cfg=agent_cfg,
                extensions=extensions,
                context=context,
            )
            if reset:
                return _fresh_selection(agent, context, agent_metadata, extensions)
        return ImageSelection(
            image_ref=agent_cfg.image_ref,
            base=base,
            extensions=list(agent_cfg.extensions),
            base_image_ref=base_image_ref(agent_metadata, agent, base, context),
        )

    if not base:
        return _fresh_selection(agent, context, agent_metadata, extensions)

    _validate_base(agent, base, agent_metadata)
    return handle_extension_selection(
        ExtensionSelectionContext(
            agent=agent,
            base=base,
            agent_cfg=agent_cfg,
            agent_metadata=agent_metadata,
            extensions=extensions,
            context=context,
        )
    )


def _require_agent_metadata(agent: str, images_metadata: ImagesMetadata) -> AgentMetadata:
    agent_metadata = images_metadata.agents.get(agent)
    if not agent_metadata:
        raise CliError(f"Agent '{agent}' is missing from images metadata.")
    return agent_metadata


def _available_bases(
    agent: str,
    agent_metadata: AgentMetadata,
) -> list[str]:
    if not agent_metadata.valid_bases:
        raise CliError(f"Agent '{agent}' does not define any valid bases.")
    return sorted(agent_metadata.valid_bases)


def _validate_base(
    agent: str,
    base: str,
    agent_metadata: AgentMetadata,
) -> None:
    if base not in agent_metadata.valid_bases:
        raise CliError(f"Base '{base}' is not valid for agent '{agent}'.")


def _fresh_selection(
    agent: str,
    context: ConfigContext,
    agent_metadata: AgentMetadata,
    extensions: dict[str, ExtensionMetadata],
) -> ImageSelection:
    available_bases = _available_bases(agent, agent_metadata)
    if not available_bases:
        raise CliError(f"No base images found for agent '{agent}' in metadata.")

    extended_images = load_extended_image_options(agent, agent_metadata, extensions)
    request = ImageChoiceRequest(
        agent=agent,
        context=context,
        agent_metadata=agent_metadata,
        extended_options=extended_images,
    )
    choice = (
        prompt_for_image_choice(request)
        if extended_images
        else ImageChoice(kind="base", value=prompt_for_base(_base_request(request)))
    )
    if choice.kind == "extended":
        selected = resolve_extended_image(choice.value, extended_images)
        return apply_extended_selection(
            agent=agent,
            agent_cfg=context.project_cfg.agents.setdefault(agent, AgentConfig()),
            selected=selected,
            agent_metadata=agent_metadata,
            context=context,
        )
    base = choice.value
    agent_cfg = context.project_cfg.agents.setdefault(agent, AgentConfig())
    agent_cfg.base = base
    return handle_extension_selection(
        ExtensionSelectionContext(
            agent=agent,
            base=base,
            agent_cfg=agent_cfg,
            agent_metadata=agent_metadata,
            extensions=extensions,
            context=context,
        )
    )


def _base_request(request: ImageChoiceRequest) -> BaseSelectionRequest:
    return BaseSelectionRequest(
        agent=request.agent,
        context=request.context,
        agent_metadata=request.agent_metadata,
    )



from __future__ import annotations

import urllib.error
import urllib.request
from collections.abc import Mapping

from ._registry_api import RegistryDiscoveryError, fetch_pull_token_for_repository
from ._timeouts import REGISTRY_REQUEST_TIMEOUT_SECONDS
from .types import RemoteImageRef


def get_remote_repo_digest(image: RemoteImageRef) -> str | None:
    reference = _parse_reference(image.image.image_ref)
    if reference is None:
        return None
    try:
        token = fetch_pull_token_for_repository(image.registry_api, image.image.repository)
    except RegistryDiscoveryError:
        return None
    url = f"{image.registry_api.registry_api_url}/{image.image.repository}/manifests/{reference}"
    headers: dict[str, str] = {
        "Accept": ",".join(
            [
                "application/vnd.oci.image.index.v1+json",
                "application/vnd.docker.distribution.manifest.list.v2+json",
                "application/vnd.oci.image.manifest.v1+json",
                "application/vnd.docker.distribution.manifest.v2+json",
            ]
        ),
        "Authorization": f"Bearer {token}",
    }
    response_headers = _head_request(url, headers)
    if response_headers is None:
        return None
    digest = response_headers.get("Docker-Content-Digest")
    if digest:
        return digest
    return response_headers.get("docker-content-digest")


def _parse_reference(image_ref: str) -> str | None:
    if "@" in image_ref:
        _, reference = image_ref.split("@", 1)
    else:
        last_colon = image_ref.rfind(":")
        if last_colon > image_ref.rfind("/"):
            reference = image_ref[last_colon + 1 :]
        else:
            return None
    if not reference:
        return None
    return reference


def _head_request(url: str, headers: Mapping[str, str]) -> dict[str, str] | None:
    request = urllib.request.Request(url, headers=dict(headers), method="HEAD")
    try:
        with urllib.request.urlopen(request, timeout=REGISTRY_REQUEST_TIMEOUT_SECONDS) as response:
            return dict(response.headers)
    except urllib.error.HTTPError as exc:
        if exc.code in {401, 403}:
            return dict(exc.headers)
        return None
    except urllib.error.URLError:
        return None



from __future__ import annotations

import os
from datetime import datetime, timezone
from pathlib import Path

from ._store import sanitize

_DEFAULT_LOG_DIR = "~/.aicage/logs/build"


def build_log_path(agent: str, base: str) -> Path:
    log_dir = Path(os.path.expanduser(_DEFAULT_LOG_DIR))
    return log_dir / f"{sanitize(agent)}-{base}-{_timestamp()}.log"


def build_log_path_for_image(image_ref: str) -> Path:
    log_dir = Path(os.path.expanduser(_DEFAULT_LOG_DIR))
    return log_dir / f"{sanitize(image_ref)}-{_timestamp()}.log"


def _timestamp() -> str:
    return datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")



from __future__ import annotations

import json
from pathlib import Path

from aicage._logging import get_logger
from aicage.docker._client import get_docker_client


def run_pull(image_ref: str, log_path: Path) -> None:
    logger = get_logger()
    log_path.parent.mkdir(parents=True, exist_ok=True)
    print(f"[aicage] Pulling image {image_ref} (logs: {log_path})...")
    logger.info("Pulling image %s (logs: %s)", image_ref, log_path)

    client = get_docker_client()
    with log_path.open("w", encoding="utf-8") as log_handle:
        for event in client.api.pull(image_ref, stream=True, decode=True):
            log_handle.write(f"{_format_pull_event(event)}\n")
            log_handle.flush()

    logger.info("Image pull succeeded for %s", image_ref)


def _format_pull_event(event: object) -> str:
    if isinstance(event, bytes):
        return event.decode("utf-8", errors="replace").rstrip("\n")
    if isinstance(event, str):
        return event.rstrip("\n")
    if isinstance(event, dict):
        return json.dumps(event, ensure_ascii=True)
    return str(event).rstrip("\n")



from dataclasses import dataclass

from aicage._logging import get_logger
from aicage.config.context import ConfigContext
from aicage.errors import CliError
from aicage.registry.images_metadata.models import AgentMetadata

from ._tty import ensure_tty_for_prompt


@dataclass(frozen=True)
class BaseSelectionRequest:
    agent: str
    context: ConfigContext
    agent_metadata: AgentMetadata


@dataclass(frozen=True)
class BaseOption:
    base: str
    description: str


def prompt_for_base(request: BaseSelectionRequest) -> str:
    ensure_tty_for_prompt()
    logger = get_logger()
    title = f"Select base image for '{request.agent}' (runtime to use inside the container):"
    bases = base_options(request.context, request.agent_metadata)

    if bases:
        print(title)
        for idx, option in enumerate(bases, start=1):
            suffix = " (default)" if option.base == request.context.global_cfg.default_image_base else ""
            print(f"  {idx}) {option.base}: {option.description}{suffix}")
        prompt = f"Enter number or name [{request.context.global_cfg.default_image_base}]: "
    else:
        prompt = f"{title} [{request.context.global_cfg.default_image_base}]: "

    response = input(prompt).strip()
    if not response:
        choice = request.context.global_cfg.default_image_base
    elif response.isdigit() and bases:
        idx = int(response)
        if idx < 1 or idx > len(bases):
            raise CliError(f"Invalid choice '{response}'. Pick a number between 1 and {len(bases)}.")
        choice = bases[idx - 1].base
    else:
        choice = response

    if bases and choice not in available_bases(bases):
        options = ", ".join(available_bases(bases))
        raise CliError(f"Invalid base '{choice}'. Valid options: {options}")
    logger.info("Selected base '%s' for agent '%s'", choice, request.agent)
    return choice


def base_options(context: ConfigContext, agent_metadata: AgentMetadata) -> list[BaseOption]:
    return [
        BaseOption(
            base=base,
            description=context.images_metadata.bases[base].base_image_description,
        )
        for base in sorted(agent_metadata.valid_bases)
    ]


def available_bases(options: list[BaseOption]) -> list[str]:
    return [option.base for option in options]



from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.runtime_config import RunConfig
from aicage.runtime.agent_config import AgentConfig, resolve_agent_config
from aicage.runtime.run_args import DockerRunArgs, merge_docker_args


def build_run_args(config: RunConfig, parsed: ParsedArgs) -> DockerRunArgs:
    agent_config: AgentConfig = resolve_agent_config(config.agent, config.images_metadata)

    merged_docker_args: str = merge_docker_args(
        config.project_docker_args,
        parsed.docker_args,
    )

    return DockerRunArgs(
        image_ref=config.image_ref,
        project_path=config.project_path,
        agent_config_host=agent_config.agent_config_host,
        agent_config_mount_container=Path("/aicage/agent-config"),
        merged_docker_args=merged_docker_args,
        agent_args=parsed.agent_args,
        agent_path=agent_config.agent_path,
        mounts=config.mounts,
    )



from __future__ import annotations

import hashlib
from typing import Protocol


class HashWriter(Protocol):
    def update(self, data: bytes, /) -> None:
        ...

    def hexdigest(self) -> str:
        ...


def new_hasher() -> HashWriter:
    return hashlib.sha256()



from __future__ import annotations

from aicage.config.runtime_config import RunConfig
from aicage.docker.build import run_build
from aicage.errors import CliError

from ._digest import refresh_base_digest
from ._logs import build_log_path
from ._plan import base_image_ref, base_repository, now_iso, should_build
from ._store import BuildRecord, BuildStore


def ensure_local_image(run_config: RunConfig) -> None:
    agent_metadata = run_config.images_metadata.agents[run_config.agent]
    if agent_metadata.local_definition_dir is None:
        raise CliError(f"Missing local definition for '{run_config.agent}'.")

    if run_config.agent_version is None:
        raise CliError(f"Missing agent version for '{run_config.agent}'.")

    base_image = base_image_ref(run_config)
    base_repo = base_repository(run_config)
    refresh_base_digest(
        base_image_ref=base_image,
        base_repository=base_repo,
        global_cfg=run_config.global_cfg,
    )

    store = BuildStore()
    record = store.load(run_config.agent, run_config.base)

    needs_build = should_build(
        run_config=run_config,
        record=record,
        base_image_ref=base_image,
    )
    if not needs_build:
        return

    log_path = build_log_path(run_config.agent, run_config.base)
    run_build(
        run_config=run_config,
        base_image_ref=base_image,
        log_path=log_path,
    )

    store.save(
        BuildRecord(
            agent=run_config.agent,
            base=run_config.base,
            agent_version=run_config.agent_version,
            base_image=base_image,
            image_ref=run_config.image_ref,
            built_at=now_iso(),
        )
    )



import argparse
import sys
from collections.abc import Sequence

from aicage._logging import get_logger
from aicage.cli_types import ParsedArgs
from aicage.errors import CliError

_MIN_REMAINING_WITH_AGENT = 2


def parse_cli(argv: Sequence[str]) -> ParsedArgs:
    """
    Returns parsed CLI args.
    Docker args are captured as an opaque string; precedence is resolved later.
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--dry-run", action="store_true", help="Print docker run command without executing.")
    parser.add_argument(
        "--aicage-entrypoint",
        dest="entrypoint",
        help="Override the container entrypoint with a host path.",
    )
    parser.add_argument("--docker", action="store_true", help="Mount the host Docker socket into the container.")
    parser.add_argument("--config", help="Perform config actions such as 'print'.")
    parser.add_argument("-h", "--help", action="store_true", help="Show help message and exit.")
    pre_argv, post_argv = _split_argv(argv)

    opts: argparse.Namespace
    remaining: list[str]
    opts, remaining = parser.parse_known_args(pre_argv)

    if opts.help:
        usage: str = (
            "Usage:\n"
            "  aicage <agent>\n"
            "  aicage [--dry-run] [--docker] [--aicage-entrypoint PATH] -- <agent> [<agent-args>]\n"
            "  aicage [--dry-run] [--docker] [--aicage-entrypoint PATH] <docker-args> -- <agent> [<agent-args>]\n"
            "  aicage --config print\n\n"
            "Any arguments between aicage and the agent require a '--' separator before the agent.\n"
            "<docker-args> are any arguments not recognized by aicage.\n"
            "These arguments are forwarded verbatim to docker run.\n"
            "<agent-args> are passed verbatim to the agent.\n"
        )
        print(usage)
        get_logger().info("Displayed CLI usage help.")
        sys.exit(0)

    if opts.config:
        _validate_config_action(opts, remaining, post_argv)
        return ParsedArgs(
            opts.dry_run,
            "",
            "",
            [],
            opts.entrypoint,
            opts.docker,
            opts.config,
        )

    docker_args, agent, agent_args = _parse_agent_section(remaining, post_argv)

    if not agent:
        raise CliError("Agent name is required.")

    return ParsedArgs(
        opts.dry_run,
        docker_args,
        agent,
        agent_args,
        opts.entrypoint,
        opts.docker,
        None,
    )


def _split_argv(argv: Sequence[str]) -> tuple[list[str], list[str] | None]:
    if "--" not in argv:
        return list(argv), None
    sep_index = argv.index("--")
    pre_argv = list(argv[:sep_index])
    post_argv = list(argv[sep_index + 1 :])
    return pre_argv, post_argv


def _validate_config_action(
    opts: argparse.Namespace,
    remaining: list[str],
    post_argv: list[str] | None,
) -> None:
    if opts.config != "print":
        raise CliError(f"Unknown config action: {opts.config}")
    if remaining or post_argv or opts.entrypoint or opts.docker or opts.dry_run:
        raise CliError("No additional arguments are allowed with --config.")


def _parse_agent_section(
    remaining: list[str],
    post_argv: list[str] | None,
) -> tuple[str, str, list[str]]:
    if post_argv is not None:
        if not post_argv:
            raise CliError("Missing agent after '--'.")
        docker_args = " ".join(remaining).strip()
        return docker_args, post_argv[0], post_argv[1:]
    if not remaining:
        raise CliError("Missing arguments. Provide an agent name (and optional docker args).")
    first: str = remaining[0]
    if first.startswith("-") or "=" in first:
        if len(remaining) < _MIN_REMAINING_WITH_AGENT:
            raise CliError("Missing agent name after docker args. Use '--' before the agent.")
        return first, remaining[1], remaining[2:]
    return "", first, remaining[1:]



from __future__ import annotations

from aicage.config.resources import find_packaged_path
from aicage.errors import CliError
from aicage.paths import IMAGES_METADATA_FILENAME
from aicage.registry._agent_discovery import discover_agents

from .models import ImagesMetadata


def load_images_metadata(local_image_repository: str) -> ImagesMetadata:
    path = find_packaged_path(IMAGES_METADATA_FILENAME)
    try:
        payload = path.read_text(encoding="utf-8")
    except OSError as exc:
        raise CliError(f"Failed to read images metadata from {path}: {exc}") from exc
    metadata = ImagesMetadata.from_yaml(payload)
    return discover_agents(metadata, local_image_repository)



from __future__ import annotations

import hashlib
import os
from pathlib import Path
from typing import Any

import yaml

from aicage.paths import CONFIG_FILENAME

from .errors import ConfigError
from .global_config import GlobalConfig
from .project_config import ProjectConfig
from .resources import find_packaged_path

_DEFAULT_BASE_DIR = "~/.aicage"
_PROJECTS_SUBDIR = "projects"


class SettingsStore:
    """
    Persists per-project configuration under ~/.aicage.
    """

    def __init__(self, base_dir: Path | None = None) -> None:
        self.base_dir = base_dir or Path(os.path.expanduser(_DEFAULT_BASE_DIR))
        self.projects_dir = self.base_dir / _PROJECTS_SUBDIR
        self.base_dir.mkdir(parents=True, exist_ok=True)
        self.projects_dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _load_yaml(path: Path) -> dict[str, Any]:
        if not path.exists():
            return {}
        try:
            with path.open("r", encoding="utf-8") as handle:
                data = yaml.safe_load(handle)
                return data or {}
        except yaml.YAMLError as exc:
            raise ConfigError(f"Failed to parse YAML config at {path}: {exc}") from exc

    @staticmethod
    def _save_yaml(path: Path, data: dict[str, Any]) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf-8") as handle:
            yaml.safe_dump(data, handle, sort_keys=True)

    def load_global(self) -> GlobalConfig:
        data = self._load_yaml(self._global_config())
        return GlobalConfig.from_mapping(data)

    def _project_path(self, project_realpath: Path) -> Path:
        digest = hashlib.sha256(str(project_realpath).encode("utf-8")).hexdigest()
        return self.projects_dir / f"{digest}.yaml"

    def load_project(self, project_realpath: Path) -> ProjectConfig:
        data = self._load_yaml(self._project_path(project_realpath))
        return ProjectConfig.from_mapping(project_realpath, data)

    def save_project(self, project_realpath: Path, config: ProjectConfig) -> None:
        self._save_yaml(self._project_path(project_realpath), config.to_mapping())


    @staticmethod
    def _global_config() -> Path:
        """
        Returns the path to the packaged global config file.
        """
        return find_packaged_path(CONFIG_FILENAME)

    def project_config_path(self, project_realpath: Path) -> Path:
        """
        Returns the path to a project's config file under the base directory.
        """
        return self._project_path(project_realpath)
